// readtest.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include <WINDOWSX.H>
#include <TCHAR.H>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glut.h>
#include "glext.h"

/////////////////////////////////////////////////////////////////////////////
// some globals

static GLsizei   iTexWidth  = 256, iTexHeight = 256;
static GLboolean bInit     = GL_TRUE;
static GLdouble  dSpeedRate = 0.0;
static GLuint    iTestType  = 0;

/////////////////////////////////////////////////////////////////////////////
// global output helpers

FILE * hF=0;
char sz[1000];

/////////////////////////////////////////////////////////////////////////////
// glReadBuffer formats 

#define NUM_FORMATS 4

struct FormatRec 
{
   int    iNum;
   GLenum nFormat;
   GLenum nType;
   GLint  iPixelSize;                                    
};

static const struct FormatRec FormatTable[NUM_FORMATS] = 
{
   { 1,  GL_RGB,      GL_UNSIGNED_BYTE, 3 },
   { 2,  GL_RGBA,     GL_UNSIGNED_BYTE, 4 },
   { 3,  GL_BGR_EXT,  GL_UNSIGNED_BYTE, 3 },
   { 4,  GL_BGRA_EXT, GL_UNSIGNED_BYTE, 4 }
};

static GLint Format;

/////////////////////////////////////////////////////////////////////////////
// main measurement func
// - setup pixel/texture stuff, creates a 256x256 texture, creates some read buffer
// - in the measurement loop: draws a textured triangle and reads a 256x256 frambuffer area

static void MeasureReadingRate(void)
{
 const int w = iTexWidth;
 const int h = iTexHeight;
 const int bytes = w * h * 4;   // 4 bytes is ok for all tests
 GLuint gTexName;
 GLubyte *texImage, *getImage;
 GLdouble t0, t1, time;
 int count,i;

 // 1 byte alignment... changing it to some WORD/DWORD/WHATEVER
 // alignment doesn't change anything on my Radeon 9700 Pro

 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glPixelStorei(GL_PACK_ALIGNMENT, 1);

 // read front or backbuffer (depends on test pass)

 if(iTestType==1) glReadBuffer(GL_FRONT);
 else             glReadBuffer(GL_BACK);

 // creates texture/read buffer

 texImage = (GLubyte *) malloc(bytes);
 memset(texImage,0x88,bytes);
 getImage = (GLubyte *) malloc(bytes);
 memset(getImage,0x99,bytes);
 if (!texImage || !getImage) 
  {
   dSpeedRate = 0.0;
   return;
  }

 // fill texture with some data

 for(i=0;i<bytes;i++) 
  {
   texImage[i] = i & 0xff;
  }

 // creates texture, setup some pixel und texture behaviour

 glGenTextures(1, &gTexName);
 glBindTexture(GL_TEXTURE_2D, gTexName);

 glPixelTransferf(GL_RED_SCALE, 1.0);
 glPixelTransferf(GL_GREEN_SCALE, 1.0);
 glPixelTransferf(GL_BLUE_SCALE, 1.0);
 glPixelTransferf(GL_RED_BIAS, 0.0);
 glPixelTransferf(GL_GREEN_BIAS, 0.0);
 glPixelTransferf(GL_BLUE_BIAS, 0.0);

 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

 glColor4ub(0xFF,0xFF,0xFF,0xFF);                      

 glTexImage2D(GL_TEXTURE_2D, 0, 3, w, h, 0,
              GL_RGB,GL_UNSIGNED_BYTE,texImage);

 glEnable(GL_TEXTURE_2D);

 // ok, main timing loop ------------------------------- //

 count = 0;
 t0 = glutGet(GLUT_ELAPSED_TIME) * 0.001;
 do 
  {
   // draw a textured triangle (just to draw something)

   glBegin(GL_TRIANGLES);

   glTexCoord2f(0, 0);     glVertex2f(200, 50);
   glTexCoord2f(1, 0);     glVertex2f(300, 50);
   glTexCoord2f(0.5, 1);   glVertex2f(250, 150);

   glEnd();

   // ok, now the read

   glReadPixels(0,0,w,h,
                FormatTable[Format].nFormat,
                FormatTable[Format].nType,
                getImage);

   // measure time

   t1 = glutGet(GLUT_ELAPSED_TIME) * 0.001;
   time = t1 - t0;
   count++;
  } while (time < 3.0);

 // ---------------------------------------------------- //

 // clean up texture

 glDisable(GL_TEXTURE_2D);

 glDeleteTextures(1, &gTexName);

 free(texImage);
 free(getImage);

 // check for error (we are using GL_BGR(A)_EXT without extension check... so at least we give some error here, if not supported... but every card I know suports that formats)

 GLint err = glGetError();
 if (err)
  {
   sprintf(sz,"ERROR %d\r\n", err);
   if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
  }

 // ok, the speedrate

 dSpeedRate = w * h * count / time;
}

/////////////////////////////////////////////////////////////////////////////
// small text output helper

static void PrintString(const char *s)
{
 while (*s) 
  {
   glutBitmapCharacter(GLUT_BITMAP_8_BY_13, (int) *s);
   s++;
  }
}

/////////////////////////////////////////////////////////////////////////////
// draw main infos here

static void DrawInfoText(void)
{
 glClearColor(0.0, 0.0, 0.0, 0.0);

 glClear(GL_COLOR_BUFFER_BIT);

 glColor4ub(0x00,0x00,0x80,0xFF);                      // full color intensity (modulate)

 glBegin(GL_QUADS);
    glVertex2f(0,   0);
    glVertex2f(0,   200);
    glVertex2f(200, 200);
    glVertex2f(200, 0);
 glEnd();

 glColor4ub(0xFF,0xFF,0xFF,0xFF);                      // full color intensity (modulate)

 glRasterPos2i(10, 280);
 PrintString("Pete's Framebuffer reading test...");

 glRasterPos2i(10, 180);
 PrintString("Test running...");

 if(!bInit)
  {
   char s[128];

   glRasterPos2i(10, 165);
   s[0]=0;
   if(iTestType==0) sprintf(s, "Backbuffer reading");
   if(iTestType==1) sprintf(s, "Frontbuffer reading");
   PrintString(s);
  
   glRasterPos2i(10, 150);
   sprintf(s, "Format: %d / %d", Format+1,NUM_FORMATS);
   PrintString(s);
  }

 glRasterPos2i(10, 10);
 PrintString("Please wait...");
}

/////////////////////////////////////////////////////////////////////////////
// glut display func... will step through the available formats as well

static void Display(void)
{
 // init the screen the first time and swap buffers

 if(bInit)
  {
   DrawInfoText();
   glutSwapBuffers();
   bInit=GL_FALSE;
  }

 // ok, draw info text

 DrawInfoText();

 // measure speed

 MeasureReadingRate();

 // create info text

 sprintf(sz,"Format: %02d - ", Format+1);
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
         
 sprintf(sz,"Speed:  %g Mpix/s  %g MB/s\r\n\r\n",
         dSpeedRate / 1000000.0,
         dSpeedRate * FormatTable[Format].iPixelSize / 1000000.0);
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);

 // step through avaible formats... if we have done all 
 // backbuffer tests: swap to frontbuffer test. After we
 // also have checked that ones, stop and show info file

 Format++;

 if(Format==NUM_FORMATS) 
  {
   Format=0;
   iTestType++;

   // backbuffer done? now frontbuffer
   if(iTestType==1)
    {
     sprintf(sz,"\r\n\r\nFrontbuffer reading speed (front)\r\n");
     if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
     sprintf(sz,"---------------------------------\r\n\r\n");
     if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
    }
   else
   // both buffers done? show info and exit
    {
     fclose(hF);
     WinExec("notepad PeteReadSpeed.txt",SW_SHOW);
     exit(0);
    }
  }

 glutSwapBuffers();
 glutPostRedisplay();
}

/////////////////////////////////////////////////////////////////////////////
// setup matrices

static void Reshape(int width, int height)
{
 glViewport( 0, 0, width, height );
 glMatrixMode( GL_PROJECTION );
 glLoadIdentity();
 glOrtho(0, width, 0, height, -1, 1);
 glMatrixMode( GL_MODELVIEW );
 glLoadIdentity();
}

/////////////////////////////////////////////////////////////////////////////
// keyboard (ESC) handler

static void Key(unsigned char key, int x, int y)
{
 switch (key) 
  {
   case 27:
      fclose(hF);
      exit(0);
      break;
  }
 glutPostRedisplay();
}

/////////////////////////////////////////////////////////////////////////////
// file init

static void Init(void)
{
 hF=fopen("PeteReadSpeed.txt","wb");

 sprintf(sz,"Pete's framebuffer read speed test\r\n");
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
 sprintf(sz,"Mail: BlackDove@addcom.de\r\n");
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
 sprintf(sz,"Homepage: http://home.t-online.de/home/PeteBernert/\r\n\r\n");
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);

 sprintf(sz,"Vendor: %s\r\n", (const char *) glGetString(GL_VENDOR));
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
 sprintf(sz,"Version: %s\r\n", (const char *) glGetString(GL_VERSION));
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
 sprintf(sz,"Renderer: %s\r\n", (const char *) glGetString(GL_RENDERER));
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
 sprintf(sz,"\r\n");
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);

 if(hF) 
  {
   sprintf(sz,"---------------------------------\r\n");
   fwrite(sz,lstrlen(sz),1,hF);
   sprintf(sz,"OpenGL extensions:\r\n");
   fwrite(sz,lstrlen(sz),1,hF);
   char * ex=(char *)glGetString(GL_EXTENSIONS);
   if(ex) fwrite(ex,lstrlen(ex),1,hF);
   sprintf(sz,"\r\n");
   fwrite(sz,lstrlen(sz),1,hF);
   sprintf(sz,"---------------------------------\r\n\r\n");
   fwrite(sz,lstrlen(sz),1,hF);
  }

 sprintf(sz,"Frontbuffer reading speed (back)\r\n");
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);

 sprintf(sz,"---------------------------------\r\n\r\n");
 if(hF) fwrite(sz,lstrlen(sz),1,hF);printf(sz);
}

/////////////////////////////////////////////////////////////////////////////
// main

int main(int argc, char *argv[])
{
 glutInit( &argc, argv );
 glutInitWindowPosition( 0, 0 );
 glutInitWindowSize( 300, 300 );
 glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
 glutCreateWindow(argv[0]);
 glutReshapeFunc( Reshape );
 glutKeyboardFunc( Key );
 glutDisplayFunc( Display );
 Init();
 glutMainLoop();
 return 0;
}

/////////////////////////////////////////////////////////////////////////////
